----------------------------------------
-- copy a table and all its subtables --
----------------------------------------
local function tcopy(intable,outtable,overwrite,written)
	--written is an internal table to track references to already copied tables and keep those intact
	--create outtable if it doesn't exist
	outtable = outtable or {}
	--create written if this is the first call (user call)
	written = written or {}
	--add copied table to written
	written[tostring(intable)] = outtable
	--loop through data
	for key, data in pairs(intable) do
		--if overwrite or not exists, write value
		if (not outtable[key]) or overwrite then
			if type(data) == "table" then
				outtable[key] = written[tostring(data)] or tcopy(data,nil,nil,written)
			else
				outtable[key] = data
			end
		end
	end
	return outtable
end

local L = LibStub("AceLocale-3.0"):GetLocale("CCAssign")
local OneRingLib = OneRingLib
local ringID
local addon = CreateFrame("Frame", "CCAssign")
local cc_db = setmetatable({}, {__index=function(t,k) t[k] = {} return t[k] end})
local cc_by_lvl_class = setmetatable({}, {__index=function(t, class)
	t[class] = setmetatable({},{__index=function(ct, lvl)
		local cctbl = {}
		for key, value in pairs(cc_db[class]) do
			if (value.l <= lvl) then
				tinsert(cctbl, value)
			end
		end
		ct[lvl] = cctbl
		return cctbl
	end})
	return t[class]
end})
local function setraidtarget(_,raidtargetid)
	SetRaidTarget("target", raidtargetid)
end
local mark_to_ringbtn ={
	[8]=1,
	[7]=2,
	[6]=4,
	[5]=8,
	[4]=7,
	[3]=6,
	[2]=5,
	[1]=3,
}
local mark_by_group_unit, group_unit_by_mark = {}, {}
local name_to_mark = {}
local cc_by_mark = {}
local uid_to_name = {}
for mark=1, 6 do cc_by_mark[mark] = {} end
local ringData={
	name="CC Assign",
	hotkey="SHIFT+C",
	[1]={"func", setraidtarget, 8, r=1,g=1,b=1,icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_8",caption=L["Kill target"],},
	[2]={"func", setraidtarget, 7, r=1,g=0,b=0,icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_7",caption=L["Secondary target"],},
	[3]={"func", setraidtarget, 1, r=1, g=1, b=0, icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_1",caption=L["Mark target with: %s"]:format(RAID_TARGET_1),},
	[4]={"func", setraidtarget, 6, r=0.20, g=0.20, b=1, icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_6",caption=L["Mark target with: %s"]:format(RAID_TARGET_6),},
	[5]={"func", setraidtarget, 2, r=1, g=0.5, b=0.05, icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_2",caption=L["Mark target with: %s"]:format(RAID_TARGET_2),},
	[6]={"func", setraidtarget, 3, r=1, g=0.30, b=1, icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_3",caption=L["Mark target with: %s"]:format(RAID_TARGET_3),},
	[7]={"func", setraidtarget, 4, r=0.20, g=1, b=0.20, icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_4",caption=L["Mark target with: %s"]:format(RAID_TARGET_4),},
	[8]={"func", setraidtarget, 5, r=0.65, g=0.84, b=1, icon="Interface\\TargetingFrame\\UI-RaidTargetingIcon_5",caption=L["Mark target with: %s"]:format(RAID_TARGET_5),},
}
local function ring_update(t)
	OneRingLib:SetRingData(ringID,tcopy(t))
	ringData = tcopy(t)
end
local eRingData = tcopy(ringData)
local sortfnc = function(a,b)
	return a.p < b.p
end
function addon:UpdateRing()
	if (GetNumRaidMembers() > 0) or (GetNumPartyMembers()<=0) or (not UnitIsEnemy("player", "target")) then ring_update(eRingData) return end
	
	local targetType = UnitCreatureType("target")
	local ringdata = ringData
	for rtID=1, 6 do
		local cc_src = uid_to_name[group_unit_by_mark[rtID]]
		local cc_name, cc_icon
		if cc_src then
			table.sort(cc_by_mark[rtID], sortfnc)
			for _, data in ipairs(cc_by_mark[rtID]) do
				if data.t[targetType] then
					if type(data.n) == "number" then
						data.n = GetSpellInfo(data.n)
					end
					cc_name, cc_icon = data.n, data.i
					break
				end
			end
		end
		local r = ringdata[mark_to_ringbtn[rtID]]
		if (cc_name and cc_icon) then
			r.icon = cc_icon
			r.caption = L["Mark for %s to %s"]:format(cc_src, cc_name)
		else
			local odata = eRingData[mark_to_ringbtn[rtID]]
			r.icon = odata.icon
			r.caption = odata.caption
		end
	end
	ring_update(ringdata)
	collectgarbage()
end
function addon:FetchGroup()
	table.wipe(mark_by_group_unit)
	table.wipe(group_unit_by_mark)
	table.wipe(cc_by_mark)
	table.wipe(uid_to_name)
	if GetNumRaidMembers() > 0 then return end
	local newmarks = {}
	local marks = {1,2,3,4,5,6}
	for i=1, GetNumPartyMembers() do
		local n = UnitName("party"..i)
		if name_to_mark[n] then
			for c, v in ipairs(marks) do
				if v == name_to_mark[n] then
					tremove(marks, c)
					break
				end
			end
			mark_by_group_unit["party"..i] = name_to_mark[n]
			group_unit_by_mark[name_to_mark[n]] = "party"..i
		else
			tinsert(newmarks, i)
		end
	end
	for _, pid in ipairs(newmarks) do
		local n = UnitName("party"..pid)
		local r = math.random(1, #marks)
		mark_by_group_unit["party"..pid] = marks[r]
		group_unit_by_mark[marks[r]] = "party"..pid
		tremove(marks, r)
	end
	table.wipe(name_to_mark)
	for pid=1, GetNumPartyMembers() do
		local n = UnitName("party"..pid)
		name_to_mark[n] = mark_by_group_unit["party"..pid]
		uid_to_name["party"..pid]=n
		cc_by_mark[mark_by_group_unit["party"..pid]] = tcopy(cc_by_lvl_class[(select(2, UnitClass("party"..pid)))][UnitLevel("party"..pid)])
		SendChatMessage(L["<CC Assign> You have been assigned raid target icon {rt%d} (%s)"]:format(mark_by_group_unit["party"..pid], _G["RAID_TARGET_"..mark_by_group_unit["party"..pid]]),"WHISPER",nil,n)
	end
	if GetNumPartyMembers() > 0 then
		SendChatMessage(L["<CC Assign> The main kill order is as follows: %s ({rt8}) over %s ({rt7}). Individual assignments for crowd control have been whispered to players."]:format(RAID_TARGET_8, RAID_TARGET_7),"PARTY")
	end
	addon:UpdateRing()
end
function addon:RegisterCC(classString, levelGained, targetTypes, texture, nameId, priority)
	print("CC registered for "..classString..", gained at level "..levelGained..", using this texture: |T"..texture..":14|t")
	print("Enemy types affected:")
	for key in pairs(targetTypes) do
		print(key)
	end
	tinsert(cc_db[classString], {l=levelGained, t=targetTypes, i=texture, n=GetSpellInfo(nameId) or nameId, p=priority})
	table.wipe(cc_by_lvl_class[classString])
end
ringID = OneRingLib:CreateRing("CCAssign", ringData)
addon:SetScript("OnEvent", function(self, event, ...)
	local arg = {...}
	if event == "ADDON_LOADED" and arg[1] == "CCAssign" then
		self:UnregisterEvent("ADDON_LOADED")
		return
	end
	if event == "PLAYER_TARGET_CHANGED" then
		self:UpdateRing()
		return
	end
	if event == "PARTY_MEMBERS_CHANGED" then
		if IsShiftKeyDown() then
			self:FetchGroup()
		end
		return
	end
end)
addon:RegisterEvent("ADDON_LOADED")
addon:RegisterEvent("PLAYER_TARGET_CHANGED")
addon:RegisterEvent("PARTY_MEMBERS_CHANGED")

SlashCmdList["CCASSIGN"]=function()
	addon:FetchGroup()
end
SLASH_CCASSIGN1 = "/ccassign"